import pytest
import os

from berea.cli import CLIConfig
from berea.bible import BibleClient
from berea.utils import get_app_data_path


def translation_exists(translation):
    return os.path.isfile(f"{get_app_data_path('translations')}/{translation}.db")


# TODO: Use a fixture or module if more helpers are needed
pytest.translation_exists = translation_exists


@pytest.fixture(scope='session', autouse=True)
def download_translations():
    for translation in ['BSB', 'KJV']:
        if not translation_exists(translation):
            bible = BibleClient(translation)
            bible.create_bible_db()
    
    default_translation = CLIConfig.get_default_translation()
    
    if default_translation != 'BSB':
        CLIConfig.set_default_translation('BSB')



import sys
import configparser
import argparse
from berea.utils import get_downloaded_translations, get_app_data_path
from berea.bible import BibleClient


# Version stored here to prevent editable install ImportError
__version__ = '0.1.0'


class CLIConfig:
    config = configparser.ConfigParser()
    path = get_app_data_path() + '/berea.ini'
    
    @classmethod
    def set_default_translation(cls, translation):
        if not cls.config.has_section('Defaults'):
            cls.config.add_section('Defaults')
        
        cls.config.set('Defaults', 'translation', translation)
        
        with open(cls.path, 'w') as config_file:
            cls.config.write(config_file)
    
    @classmethod
    def get_default_translation(cls):
        cls.config.read(cls.path)
        return cls.config.get('Defaults', 'translation', fallback=None)


def add_download_parser(subparsers):
    download_parser = subparsers.add_parser(
        'download',
        help="Download a Bible translation"
    )
    
    download_parser.add_argument(
        'translation',
        default='KJV'
    )
    
    
def add_delete_parser(subparsers, downloaded_translations):
    delete_parser = subparsers.add_parser(
        'delete',
        help="Delete a Bible translation"
    )
    
    delete_parser.add_argument(
        'translation',
        choices=downloaded_translations
    )


def add_config_parser(subparsers, downloaded_translations):
    config_parser = subparsers.add_parser(
        'config',
        help="Configure default settings"
    )

    config_parser.add_argument(
        'parameter',
        choices=['translation']
    )

    config_parser.add_argument(
        'value',
        choices=downloaded_translations
    )


def add_reference_parser(subparsers, downloaded_translations):
    reference_parser = subparsers.add_parser(
        'reference',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Reference a passage from the Bible (default command)"
    )
    
    reference_parser.add_argument(
        'book'
    )

    reference_parser.add_argument('chapter', nargs='?')
    reference_parser.add_argument('verse', nargs='?')
    
    reference_parser.add_argument(
        '-t', '--translation',
        choices=downloaded_translations,
        default=CLIConfig.get_default_translation(),
        help='Bible translation used to display passage'
    )
    
    reference_parser.add_argument(
        '-n', '--verse_numbers',
        action='store_true'
    )
    
    # TODO: Make the default format configurable
    reference_parser.add_argument(
        '-f', '--format',
        choices=['txt', 'md'],
        default='txt'
    )
    
    
def add_search_parser(subparsers, downloaded_translations):
    search_parser = subparsers.add_parser(
        'search',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Search for a specific phrase in a Bible translation"
    )
    
    search_parser.add_argument(
        'phrase',
        help="Phrase to search"
    )
    
    search_parser.add_argument('book', nargs='?')
    search_parser.add_argument('chapter', nargs='?')
    
    search_parser.add_argument(
        '-t', '--translation',
        choices=downloaded_translations,
        default=CLIConfig.get_default_translation(),
        help='Bible translation used to search phrase'
    )

    search_parser.add_argument(
        '-NT', '--new_testament',
        action='store_true'
    )

    search_parser.add_argument(
        '-OT', '--old_testament',
        action='store_true'
    )


def parse_berea_args(downloaded_translations):
    description = "Berea: A CLI for studying Scripture."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    
    subparsers = parser.add_subparsers(title="Commands", dest="command")
    add_download_parser(subparsers)
    add_delete_parser(subparsers, downloaded_translations)
    add_config_parser(subparsers, downloaded_translations)
    add_reference_parser(subparsers, downloaded_translations)
    add_search_parser(subparsers, downloaded_translations)
    
    return parser.parse_args()


def main():
    if len(sys.argv) < 2:
        sys.argv = ['bible', '--help']
    
    commands = [
        '--help',
        '-h',
        '--version',
        'download',
        'delete',
        'config',
        'reference',
        'search',
    ]
    
    # Set reference as the default command
    if sys.argv[1] not in commands:
        sys.argv.insert(1, 'reference')
    
    downloaded_translations = get_downloaded_translations()
    args = parse_berea_args(downloaded_translations)

    if args.command == 'config':
        CLIConfig.set_default_translation(args.value)
        print('Default translation updated.')
        return 

    bible = BibleClient(args.translation)
    
    if args.command == 'download':
        # Save first downloaded translation as the default
        if not downloaded_translations:
            CLIConfig.set_default_translation(args.translation)
        
        bible.create_bible_db()
    
    elif not downloaded_translations:
        print(f"Error: Download a translation before invoking '{args.command}'.")
        
    elif args.command == 'delete':
        bible.delete_translation()

        # Update config if no other translation is downloaded
        if [args.translation] == downloaded_translations:
            CLIConfig.set_default_translation('None')
        
        # Update config if default translation is deleted
        if args.translation == CLIConfig.get_default_translation():
            CLIConfig.set_default_translation(get_downloaded_translations()[0])

    elif args.command ==  'reference': 
        if not args.chapter:
            bible.print_book(args.book, args.format, args.verse_numbers)
        elif not args.verse:
            bible.print_chapter(
                args.book,
                args.chapter,
                args.format,
                args.verse_numbers
                )
        elif '-' in args.verse:
            bible.print_verses(
                args.book,
                args.chapter,
                args.verse,
                args.format,
                args.verse_numbers
            )
        else:
            bible.print_verse(
                args.book,
                args.chapter,
                args.verse,
                args.format,
                args.verse_numbers
                )
        
    elif args.command ==  'search':
        if args.chapter:
            if args.new_testament:
                print(
                    "Invalid search: cannot search a passage with the "
                    "'-NT, --new_testament' flag."
                )
            elif args.old_testament:
                print(
                    "Invalid search: cannot search a passage with the "
                    "'-OT, --old_testament' flag."
                )
            else:
                bible.search_chapter(args.phrase, args.book, args.chapter)
        elif args.book:
            if args.new_testament:
                print(
                    "Invalid search: cannot search a passage with the "
                    "'-NT, --new_testament' flag."
                )
            elif args.old_testament:
                print(
                    "Invalid search: cannot search a passage with the "
                    "'-OT, --old_testament' flag."
                )
            else:
                bible.search_book(args.phrase, args.book)
        elif args.new_testament:
            bible.search_testament(args.phrase , 'nt')
        elif args.old_testament:
            bible.search_testament(args.phrase , 'ot')
        else:
            bible.search_bible(args.phrase)


if __name__ == "__main__":
    main()



import urllib.request

import pytest

from berea.bible import list_multiline_verse, BibleClient


@pytest.mark.parametrize(
    "verse, verse_list",
    [   
        # John 3:16 KJV (141 characters)
        (
            "For God so loved the world, that he gave his only begotten Son, that whosoever believeth in him should not perish, but have everlasting life.",
            [
                "For God so loved the world, that he gave his only begotten Son, that whosoever",
                "believeth in him should not perish, but have everlasting life."
            ]
        ),
        # Esther 8:9 KJV (528 characters)
        (
            "Then were the king’s scribes called at that time in the third month, that is, the month Sivan, on the three and twentieth day thereof; and it was written according to all that Mordecai commanded unto the Jews, and to the lieutenants, and the deputies and rulers of the provinces which are from India unto Ethiopia, an hundred twenty and seven provinces, unto every province according to the writing thereof, and unto every people after their language, and to the Jews according to their writing, and according to their language.",
            [
                "Then were the king’s scribes called at that time in the third month, that is,",
                "the month Sivan, on the three and twentieth day thereof; and it was written",
                "according to all that Mordecai commanded unto the Jews, and to the",
                "lieutenants, and the deputies and rulers of the provinces which are from India",
                "unto Ethiopia, an hundred twenty and seven provinces, unto every province",
                "according to the writing thereof, and unto every people after their language,",
                "and to the Jews according to their writing, and according to their language."
            ]
        ),
    ]
)
def test_list_multiline_verse(verse, verse_list):
    assert list_multiline_verse(verse) == verse_list


def valid_url(url):
    try:
        with urllib.request.urlopen(url) as response:
            if response.status == 200:
                return True
            else:
                return False
    except urllib.error.HTTPError as e:
        print("HTTP error:", e.code)
        return False
    except urllib.error.URLError as e:
        print("URL error:", e.reason)
        return False


@pytest.mark.parametrize(
    "msg, book, chapter, verse, translation, expected_link",
    [
        (
            "Creating link for a single verse failed",
            "John", "3", "16", "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=john.3.16&options=NVHUG"
        ),
        (
            "Creating link for multiple verses failed",
            "John", "3", "16-18", "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=john.3.16-john.3.18&options=NVHUG"
        ),
        (
            "Creating link for a chapter failed",
            "Psalms", "117", None, "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=psalm.117&options=NVHUG"
        ),
        (
            "Creating link for a book",
            "III John", None, None, "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=3john&options=NVHUG"
        ),
    ]
)
def test_create_link(msg, book, chapter, verse, translation, expected_link):
    bible = BibleClient(translation)
    actual_link = bible.create_link(book, chapter, verse)
    assert actual_link == expected_link, msg
    assert valid_url(actual_link)


def test_validate_resource_abbreviations():
    """Creating links from the resource's book abbreviations yields valid URLs."""
    bible = BibleClient('BSB')
    
    books = bible.get_bible_cursor().execute("SELECT * FROM books").fetchall()
    
    for book in books:
        link = bible.create_link(book['name'])
        assert valid_url(link), f"{book['name']} produced invalid link: {link}"


@pytest.mark.parametrize(
    "translation",
    [
        ('KJV'),
        ('BSB'),
        ('RLT'),        
        ('UKJV'),        
    ]
)
def test_create_bible_db(translation):
    bible = BibleClient(translation)
    bible.create_bible_db()

    cursor = bible.get_bible_cursor()

    msg = 'Downloading the translation database failed.'
    assert pytest.translation_exists(translation), msg

    sql = "SELECT name FROM sqlite_master WHERE type='table';"
    actual_tables = [row['name'] for row in cursor.execute(sql).fetchall()]

    table_record_counts = {
        'books': 66,
        'verses': 31102,
    }

    msg = 'Renaming the database tables failed.'
    renamed_tables = table_record_counts.keys()
    assert set(renamed_tables).issubset(actual_tables), msg

    for table, expected_records_count in table_record_counts.items():
        sql = f"SELECT COUNT(*) FROM {table};"
        actual_records_count = cursor.execute(sql).fetchone()[0]
        
        msg = f"'{table}' table does not contain expected record count."
        assert actual_records_count == expected_records_count, msg

    created_tables = ['abbreviations', 'resources', 'resources_abbreviations']
    for expected_table in created_tables:
        msg =  f"'{expected_table}' table does not exist."
        assert expected_table in actual_tables, msg



import pytest
import sys

from berea.cli import main, CLIConfig
from berea.utils import get_downloaded_translations


@pytest.mark.parametrize(
    "msg, args, output", 
    [
        (
            "Printing a single verse failed",
            ['john', '3', '16'],
            (
                "For God so loved the world that He gave His one and only Son, that everyone\n"
                "who believes in Him shall not perish but have eternal life. "
            )
        ),
        (
            "Printing multiple verses failed",
            ['john', '3', '16-18'],
            (
                "For God so loved the world that He gave His one and only Son, that everyone\n"
                "who believes in Him shall not perish but have eternal life. For God did not\n"
                "send His Son into the world to condemn the world, but to save the world\n"
                "through Him. Whoever believes in Him is not condemned, but whoever does not\n"
                "believe has already been condemned, because he has not believed in the name of\n"
                "God’s one and only Son. "
            )
        ),
        (
            "Printing multiple verses with verse numbers failed",
            ['john', '3', '16-18', '-n'],
            (
                "16 For God so loved the world that He gave His one and only Son, that everyone\n"
                "who believes in Him shall not perish but have eternal life. 17 For God did not\n"
                "send His Son into the world to condemn the world, but to save the world\n"
                "through Him. 18 Whoever believes in Him is not condemned, but whoever does not\n"
                "believe has already been condemned, because he has not believed in the name of\n"
                "God’s one and only Son. "
            )
        ),
        (
            "Printing a chapter failed",
            ['psa', '117'],
            (
                "Praise the LORD, all you nations! Extol Him, all you peoples! For great is His\n"
                "loving devotion toward us, and the faithfulness of the LORD endures forever.  \n"
                "Hallelujah! "
            )
        ),
        (
            "Printing a chapter with verse numbers failed",
            ['psa', '117', '-n'],
            (
                "1 Praise the LORD, all you nations! Extol Him, all you peoples! 2 For great is\n"
                "His loving devotion toward us, and the faithfulness of the LORD endures\n"
                "forever.   Hallelujah! "
            )
        ),
        (
            "Printing a book failed",
            ['3john'],
            (
                "The elder,  To the beloved Gaius, whom I love in the truth: Beloved, I pray\n"
                "that in every way you may prosper and enjoy good health, as your soul also\n"
                "prospers. For I was overjoyed when the brothers came and testified about your\n"
                "devotion to the truth, in which you continue to walk. I have no greater joy\n"
                "than to hear that my children are walking in the truth. Beloved, you are\n"
                "faithful in what you are doing for the brothers, and especially since they are\n"
                "strangers to you. They have testified to the church about your love. You will\n"
                "do well to send them on their way in a manner worthy of God. For they went out\n"
                "on behalf of the Name, accepting nothing from the Gentiles. Therefore we ought\n"
                "to support such men, so that we may be fellow workers for the truth. I have\n"
                "written to the church about this, but Diotrephes, who loves to be first, will\n"
                "not accept our instruction. So if I come, I will call attention to his\n"
                "malicious slander against us. And unsatisfied with that, he refuses to welcome\n"
                "the brothers and forbids those who want to do so, even putting them out of the\n"
                "church. Beloved, do not imitate what is evil, but what is good. The one who\n"
                "does good is of God; the one who does evil has not seen God. Demetrius has\n"
                "received a good testimony from everyone, and from the truth itself. We also\n"
                "testify for him, and you know that our testimony is true. I have many things\n"
                "to write to you, but I would prefer not to do so with pen and ink. Instead, I\n"
                "hope to see you soon and speak with you face to face.  Peace to you.  The\n"
                "friends here send you greetings.  Greet each of our friends there by name. "
            )
        ),
        (
            "Printing a book with verse numbers failed",
            ['3john', '-n'],
            (
                "1 The elder,  To the beloved Gaius, whom I love in the truth: 2 Beloved, I\n"
                "pray that in every way you may prosper and enjoy good health, as your soul\n"
                "also prospers. 3 For I was overjoyed when the brothers came and testified\n"
                "about your devotion to the truth, in which you continue to walk. 4 I have no\n"
                "greater joy than to hear that my children are walking in the truth. 5 Beloved,\n"
                "you are faithful in what you are doing for the brothers, and especially since\n"
                "they are strangers to you. 6 They have testified to the church about your\n"
                "love. You will do well to send them on their way in a manner worthy of God. 7\n"
                "For they went out on behalf of the Name, accepting nothing from the Gentiles.\n"
                "8 Therefore we ought to support such men, so that we may be fellow workers for\n"
                "the truth. 9 I have written to the church about this, but Diotrephes, who\n"
                "loves to be first, will not accept our instruction. 10 So if I come, I will\n"
                "call attention to his malicious slander against us. And unsatisfied with that,\n"
                "he refuses to welcome the brothers and forbids those who want to do so, even\n"
                "putting them out of the church. 11 Beloved, do not imitate what is evil, but\n"
                "what is good. The one who does good is of God; the one who does evil has not\n"
                "seen God. 12 Demetrius has received a good testimony from everyone, and from\n"
                "the truth itself. We also testify for him, and you know that our testimony is\n"
                "true. 13 I have many things to write to you, but I would prefer not to do so\n"
                "with pen and ink. 14 Instead, I hope to see you soon and speak with you face\n"
                "to face.  Peace to you.  The friends here send you greetings.  Greet each of\n"
                "our friends there by name. "
            )
        ),
        (
            "Printing a different translation failed",
            ['john', '3', '16', '-t', 'KJV'],
            (
                "For God so loved the world, that he gave his only begotten Son, that whosoever\n"
                "believeth in him should not perish, but have everlasting life. "
            )    
        ),
        (
            "Creating a Markdown excerpt failed",
            ['john', '3', '16', '-f', 'md'],
            (
                "###\n\n"
                "______________________________________________________________________\n\n"
                "For God so loved the world that He gave His one and only Son, that everyone\n"
                "who believes in Him shall not perish but have eternal life. \n"
                "([John 3:16 BSB](https://www.stepbible.org/?q=version=BSB@reference=john.3.16&options=NVHUG))\n\n"
                "______________________________________________________________________"
            )
        ),
        # The BSB contains empty verses for cross-references to verses present
        # in some manuscripts
        (
            "Skipping empty verses failed",
            ['mark', '9', '43-47', '-n'],
            (
                "43 If your hand causes you to sin, cut it off. It is better for you to enter\n"
                "life crippled than to have two hands and go into hell, into the unquenchable\n"
                "fire. 45 If your foot causes you to sin, cut it off. It is better for you to\n"
                "enter life lame than to have two feet and be thrown into hell. 47 And if your\n"
                "eye causes you to sin, pluck it out. It is better for you to enter the kingdom\n"
                "of God with one eye than to have two eyes and be thrown into hell, "
            )
        )
    ]
)
def test_reference(monkeypatch, capsys, msg, args, output):
    args.insert(0, 'bible')
    monkeypatch.setattr(sys, 'argv', args)
    
    main()
    
    captured = capsys.readouterr()
    assert captured.out == output + '\n', msg


def test_delete(monkeypatch):
    default_translation = CLIConfig.get_default_translation()
    monkeypatch.setattr(sys, 'argv', ['bible', 'delete', default_translation])
    main()
    
    assert not pytest.translation_exists(default_translation)
    
    downloaded_translations = get_downloaded_translations()
    
    msg = "Failed to update config after deleting default translation"
    new_default_translation = CLIConfig.get_default_translation()
    assert new_default_translation == downloaded_translations[0], msg
    
    # Delete remaining translations
    for translation in downloaded_translations:
        monkeypatch.setattr(sys, 'argv', ['bible', 'delete', translation])
        main()
    
        assert not pytest.translation_exists(translation)
    
    msg = "Failed to update config if no other translation is downloaded"
    assert CLIConfig.get_default_translation() == 'None', msg


def test_download_sets_default_translation(monkeypatch):
    # Delete remaining translations if any exist
    downloaded_translations = get_downloaded_translations()
    if downloaded_translations:
        for translation in downloaded_translations:
            monkeypatch.setattr(sys, 'argv', ['bible', 'delete', translation])
            main()
    
    monkeypatch.setattr(sys, 'argv', ['bible', 'download', 'BSB'])
    main()
    
    msg = "Failed to set first download as the default translation"
    assert CLIConfig.get_default_translation() == 'BSB', msg


@pytest.mark.parametrize(
    "translation",
    [
        ('KJV'),
        ('BSB'),
        ('LEB'),
        ('RLT'),        
        ('UKJV'),        
    ]
)
def test_download(monkeypatch, translation):
    if translation in get_downloaded_translations():
        monkeypatch.setattr(sys, 'argv', ['bible', 'delete', translation])
        main()
    
    args = ['bible', 'download']
    args.append(translation)
    monkeypatch.setattr(sys, 'argv', args)
    
    main()
    
    assert pytest.translation_exists(translation)


@pytest.mark.parametrize(
    "msg, args, output",
    [
        (
            "Searching a phrase with a single occurrence failed",
            ['rescue my soul', '-t', 'BSB'],
            (
                "1 occurrences of 'rescue my soul' in the BSB Bible:\n"
                "___\n\n"
                "Psalms 35:17:\n"
                "How long, O Lord, will You look on? Rescue my soul from their ravages, my precious life from these lions. \n"
                "___\n"
            )
        ),
        (
            "Searching a phrase with multiple occurrences failed",
            ['sheep gate', '-t', 'BSB'],
            (
                "4 occurrences of 'sheep gate' in the BSB Bible:\n"
                "___\n\n"
                "Nehemiah 3:1:\n"
                "At the Sheep Gate, Eliashib the high priest and his fellow priests began rebuilding. They dedicated it and installed its doors. After building as far as the Tower of the Hundred and the Tower of Hananel, they dedicated the wall. \n"
                "___\n\n"
                "Nehemiah 3:32:\n"
                "And between the upper room above the corner and the Sheep Gate, the goldsmiths and merchants made repairs. \n"
                "___\n\n"
                "Nehemiah 12:39:\n"
                "over the Gate of Ephraim, the Jeshanah Gate, the Fish Gate, the Tower of Hananel, and the Tower of the Hundred, as far as the Sheep Gate. And they stopped at the Gate of the Guard. \n"
                "___\n\n"
                "John 5:2:\n"
                "Now there is in Jerusalem near the Sheep Gate a pool with five covered colonnades, which in Hebrew is called Bethesda. \n"
                "___\n"
            )
        ),
        (
            "Searching a phrase in the Old Testament failed",
            ['holy spirit', '-OT', '-t', 'BSB'],
            (
                "3 occurrences of 'holy spirit' in the Old Testament (BSB):\n"
                "___\n\n"
                "Psalms 51:11:\n"
                "Cast me not away from Your presence; take not Your Holy Spirit from me. \n"
                "___\n\n"
                "Isaiah 63:10:\n"
                "But they rebelled and grieved His Holy Spirit. So He turned and became their enemy, and He Himself fought against them.   \n"
                "___\n\n"
                "Isaiah 63:11:\n"
                "Then His people remembered the days of old, the days of Moses. Where is He who brought them through the sea with the shepherds of His flock? Where is the One who set His Holy Spirit among them, \n"
                "___\n"
            )
        ),
        (
            "Searching a phrase in the New Testament failed",
            ['ghost', '-NT', '-t', 'BSB'],
            (
                "2 occurrences of 'ghost' in the New Testament (BSB):\n"
                "___\n\n"
                "Matthew 14:26:\n"
                "When the disciples saw Him walking on the sea, they were terrified. “It’s a ghost!” they said, and cried out in fear. \n"
                "___\n\n"
                "Mark 6:49:\n"
                "but when they saw Him walking on the sea, they cried out, thinking He was a ghost— \n"
                "___\n"
            )
        ),
        (
            "Searching a phrase in a book failed",
            ['lying spirit', '2chr', '-t', 'BSB'],
            (
                "2 occurrences of 'lying spirit' in II Chronicles (BSB):\n"
                "___\n\n"
                "II Chronicles 18:21:\n"
                "And he replied, ‘I will go out and be a lying spirit in the mouths of all his prophets.’  ‘You will surely entice him and prevail,’ said the LORD. ‘Go and do it.’ \n"
                "___\n\n"
                "II Chronicles 18:22:\n"
                "So you see, the LORD has put a lying spirit in the mouths of these prophets of yours, and the LORD has pronounced disaster against you.” \n"
                "___\n"
            )
        ),
        (
            "Searching a phrase in a chapter failed",
            ['lampstands', 'rev', '1', '-t', 'BSB'],
            (
                "3 occurrences of 'lampstands' in Revelation of John 1 (BSB):\n"
                "___\n\n"
                "Revelation of John 1:12:\n"
                "Then I turned to see the voice that was speaking with me. And having turned, I saw seven golden lampstands, \n"
                "___\n\n"
                "Revelation of John 1:13:\n"
                "and among the lampstands was One like the Son of Man, dressed in a long robe, with a golden sash around His chest. \n"
                "___\n\n"
                "Revelation of John 1:20:\n"
                "This is the mystery of the seven stars you saw in My right hand and of the seven golden lampstands: The seven stars are the angels of the seven churches, and the seven lampstands are the seven churches. \n"
                "___\n"
            )
        ),
        (
            "Searching a phrase in a passage with the New Testament flag should be invalid",
            ['lampstands', 'rev', '1', '-NT'],
            (
                "Invalid search: cannot search a passage with the '-NT, --new_testament' flag."
            )
        ),
        (
            "Searching a phrase in a passage with the Old Testament flag should be invalid",
            ['holy spirit', 'isa', '-OT'],
            (
                "Invalid search: cannot search a passage with the '-OT, --old_testament' flag."
            )
        ),
    ]
)
def test_search(monkeypatch, capsys, msg, args, output):
    args = ['bible', 'search'] +  args
    monkeypatch.setattr(sys, 'argv', args)
    
    main()
    
    captured = capsys.readouterr()
    assert captured.out == output + '\n', msg


@pytest.mark.parametrize(
    "translation",
    [
        ('KJV'),
        ('LEB'),
        ('RLT'),        
        ('UKJV'),      
        ('BSB'), 
    ]
)
def test_config_default_translation(monkeypatch, translation):
    args = ['bible', 'config', 'translation', translation]
    monkeypatch.setattr(sys, 'argv', args)
    
    main()
    
    msg = "Setting the default translation failed"
    assert CLIConfig.get_default_translation() == translation, msg



import sys
from site import getsitepackages
import os


def get_source_root():
    return os.path.realpath(os.path.dirname(__file__))


def get_app_data_path(subdir=None):
    """
    Retrieve the venv or OS's directory for mutuable app data (translations or config).
    """
    app = 'berea'
    
    # Check if a virtual environment is active
    if hasattr(sys, 'prefix') and sys.prefix != sys.base_prefix:
        # Get path to venv/lib/python3.XX/site-packages/
        venv_site_packages = getsitepackages()[0]
        path = os.path.join(venv_site_packages, app)
    
    # No venv, use OS's standard path for config
    else:
        system_platform = sys.platform
        
        if system_platform == 'win32':
            path = os.path.join(os.environ.get('APPDATA', ''), app)
        elif system_platform == 'darwin':  # macOS
            path = os.path.join(os.path.expanduser('~'), 'Library', 'Application Support', app)
        elif system_platform == 'linux':
            path = os.path.join(os.path.expanduser('~'), '.config', app)
        
        else:
            sys.exit(f"Unsupported platform: {system_platform}")
    
    if subdir:
            path = os.path.join(path, subdir)
    
    # Create berea directory if it doesn't exist
    if not os.path.exists(path):
        os.makedirs(path)
        
    return path


def get_downloaded_translations():
    translations_path = get_app_data_path('translations')
    
    if os.path.exists(translations_path):
        files = os.listdir(translations_path)
        
        downloaded_translations = []
        
        for file in files:
            if file.endswith('.db'):
                downloaded_translations.append(file[:-3])
        
        return sorted(downloaded_translations)



import sqlite3
import urllib.request
from urllib.error import HTTPError
import json
import csv
import sys
import os

from berea.utils import get_source_root, get_app_data_path


def clean_book_name(book):
        cleaned_name = book.title()
        # Database names use proper title case
        # eg. 'Song of Solomon' and 'Revelation of John'
        if 'Of' in cleaned_name:
            cleaned_name = cleaned_name.replace('Of', 'of')
        return cleaned_name


def import_resource_books(resource='step_bible'):
    books = []
    
    with open(f'{get_source_root()}/data/{resource}_books.csv') as csv_file:
        csv_reader = csv.DictReader(csv_file)
        for row in csv_reader:
            books.append(row['abbreviation'])
    
    return books


def list_multiline_verse(verse):
    lines = []
    
    next_line = verse
    
    while len(next_line) > 80:
        # Split the verse if there's more than one line left
        space_split = next_line[:79].rfind(' ')
        lines.append(next_line[:space_split])
        next_line = next_line[space_split:].lstrip()
            
    # Append last line of verse
    lines.append(next_line)
    
    return lines


def parse_verses_str(verses):
    verses_split = verses.split('-')
    return verses_split[0], verses_split[1]


def list_to_sql(data):
    return "('" + "','".join(data) + "')"


class BibleClient:
    def __init__(self, translation):
        self.translation = translation
        # Use venv path or platform app data path to store translation DBs
        self.database = f"{get_app_data_path('translations')}/{self.translation}.db"
    
    def download_raw_bible(self):
        url = f"https://github.com/scrollmapper/bible_databases/raw/refs/heads/master/formats/sqlite/{self.translation}.db"

        try:
            urllib.request.urlretrieve(url, self.database)
            print(f"Downloaded: {self.database}")
            
        except HTTPError:
            link = "https://github.com/scrollmapper/bible_databases?tab=readme-ov-file#available-translations-140"
            msg = (
                f"Translation '{self.translation}' does not exist.\n"
                f"Check the following link for available translations:\n{link}"
            )
            sys.exit(msg)
    
    # TODO: Close out the conn when it's released
    def get_bible_cursor(self):
        conn = sqlite3.connect(self.database)
        conn.row_factory = sqlite3.Row
        # TODO: Use context manager?
        return conn.cursor()
    
    def rename_tables(self):
        """Rename tables for consistent schema across downloaded translations.
        """
        cursor = self.get_bible_cursor()
        
        tables = ['books', 'verses']
        for table in tables:
            # SQLite doesn't bind parameters for schema objects
            sql = f"ALTER TABLE {self.translation}_{table} RENAME TO {table};"
            cursor.execute(sql)
    
    def create_abbreviations_table(self):
        cursor = self.get_bible_cursor()

        cursor.execute(f"""
        CREATE TABLE IF NOT EXISTS abbreviations (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            book_id INTEGER,
            abbreviation TEXT,
            FOREIGN KEY (book_id) REFERENCES books(id)
        );
        """)
        
        books_to_abbreviations = {}
        
        with open(f'{get_source_root()}/data/book_abbreviations.json') as file:
            books_to_abbreviations = dict(json.load(file))
    
        # Create a conn to commit inserts and close 
        conn = sqlite3.connect(self.database)
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        
        for book, abbreviations in books_to_abbreviations.items():
            for abbreviation in abbreviations:
                params = {
                    'abbreviation': abbreviation,
                    'book': book,
                }
                
                cursor.execute(f"""
                INSERT INTO abbreviations (abbreviation, book_id)
                SELECT :abbreviation, books.id
                FROM books
                WHERE books.name = :book;
                """, params)
        
        conn.commit()
        conn.close()

    def create_resource_tables(self):
        cursor = self.get_bible_cursor()

        cursor.execute(f"""
        CREATE TABLE IF NOT EXISTS resources (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT
        );
        """)
        
        cursor.execute(f"""
        CREATE TABLE IF NOT EXISTS resources_abbreviations (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            resource_id INTEGER,
            abbreviation_id INTEGER
        );
        """)
        
        # Create a conn to commit inserts and close 
        conn = sqlite3.connect(self.database)
        cursor = conn.cursor()
        
        # TODO: Insert STEP Bible dynamically
        resource='STEP Bible'
        cursor.execute(f"""
        INSERT INTO resources (name) VALUES (
            'STEP Bible'
        );
        """)
        
        conn.commit()
        conn.close()
        
        abbreviations = import_resource_books()
        
        conn = sqlite3.connect(self.database)
        cursor = conn.cursor()
        
        for abbreviation in abbreviations:
            params = {
                'abbreviation': abbreviation.lower(),
            }
            
            # TODO: Select STEP Bible id dynamically
            cursor.execute(f"""
            INSERT INTO resources_abbreviations (resource_id, abbreviation_id)
            SELECT 1, abbreviations.id
            FROM abbreviations
            WHERE abbreviations.abbreviation = :abbreviation;
            """, params)
        
        conn.commit()
        conn.close()
    
    def create_bible_db(self):
        self.download_raw_bible()
        self.rename_tables()
        self.create_abbreviations_table()
        self.create_resource_tables()
    
    def delete_translation(self):
        os.remove(self.database)
        print(f"Deleted transation '{self.translation}'.")
    
    def create_link_label(self, book, chapter=None, verse=None):
        """Creates a link label, eg. `Isaiah 14:12-20`
        """
        label = book
        
        if chapter:
            label += f" {chapter}"
            
            if verse:
                label += f":{verse}"
        
        label += f" {self.translation}"
        
        return label
    
    def get_book_abbreviation_by_resource(self, book, resource):
        """Get a book's abbreviation used by a specific resource.
        """
        cursor = self.get_bible_cursor()
   
        params = {
            'book': book,
            'resource': resource,
        }
        
        cursor.execute("""
            SELECT abbreviation FROM abbreviations
            JOIN books ON abbreviations.book_id = books.id
            JOIN resources_abbreviations ON resources_abbreviations.abbreviation_id = abbreviations.id
            JOIN resources ON resources_abbreviations.resource_id = resources.id
            WHERE books.name = :book
            AND resources.name = :resource
            """, params)
        
        # Assuming a resource only has one abbreviation for a given book and translation
        return cursor.fetchone()[0]
    
    def get_book_from_abbreviation(self, book):
        cleaned_book_name = clean_book_name(book)
        cursor = self.get_bible_cursor()
        
        # Use full book name if that was passed in
        book_row = cursor.execute(
            """SELECT * FROM books WHERE books.name = ?;""",
            (cleaned_book_name,)).fetchone()
        
        if book_row:
            return cleaned_book_name
        
        # Get full book name using abbreviation
        else:
            book_row = cursor.execute("""
            SELECT * FROM books
            JOIN abbreviations ON abbreviations.book_id = books.id
            WHERE abbreviations.abbreviation = ?;
            """, (book,)).fetchone()
            
            if book_row:
                return book_row['name']
            else:
                msg = f"Invalid input {book=}."
                sys.exit(msg)
    
    # TODO: Link format depends on resource
    def create_link(self, book, chapter=None, verse=None, resource='STEP Bible'):
        book_abbrev = self.get_book_abbreviation_by_resource(book, resource)
        
        link = ''
        
        if verse:
            # Parse verses if multiple provided
            if '-' in verse:
                verse_start, verse_end = parse_verses_str(verse)
                link = f"https://www.stepbible.org/?q=version={self.translation}@reference={book_abbrev}.{chapter}.{verse_start}-{book_abbrev}.{chapter}.{verse_end}&options=NVHUG"
                
            else:
                link = f"https://www.stepbible.org/?q=version={self.translation}@reference={book_abbrev}.{chapter}.{verse}&options=NVHUG"
        
        elif chapter:
            link = f"https://www.stepbible.org/?q=version={self.translation}@reference={book_abbrev}.{chapter}&options=NVHUG"
        
        # Make link for whole book
        else:
            link = f"https://www.stepbible.org/?q=version={self.translation}@reference={book_abbrev}&options=NVHUG"

        return link

    # TODO: Adjustable line length? (BSB wraps lines at 40-43 characters)
    def print_single_or_multiline_verse(verse_record):
        # Print single line verse
        if len(verse_record[0]) <= 80:
            print(verse_record[0])
        
        # Split the verse into multiple lines if it's too long
        else:
            lines = list_multiline_verse(verse_record[0])
            for line in lines:
                print(line)

    # TODO: Replace consecutive spaces with single spaces
    # TODO: Input line length?
    def print_wall_of_text(self, verse_records, verse_numbers=False): 
        verses = ''
        for row in verse_records:
            # Skip empty verses so orphaned verse numbers or extra whitespace
            # is not displayed
            if not row['text']:
                continue
            if verse_numbers:
                verses += str(row['verse']) + ' '
            
            verses += row['text'].strip() + ' '
        
        verses_split = list_multiline_verse(verses)
        wrapped_verses = '\n'.join(verses_split)
        
        print(wrapped_verses)

    # TODO: Print paragraphs from Bible format
    def print_markdown_excerpt(self, verse_records, book, chapter, verse):
        """Generate Markdown excerpt for the verses.

        Args:
            verse_records (_type_): _description_
            params (_type_): _description_
        """
        print('###\n')
        print('______________________________________________________________________\n')
        self.print_wall_of_text(verse_records)
        print(
            f"([{self.create_link_label(book, chapter, verse,)}]"
            f"({self.create_link(book, chapter, verse,)}))"
        )
        print('\n______________________________________________________________________')

    # TODO: Print paragraphs from Bible format
    def print_passage_by_format(self, format, verse_records, verse_numbers=False, book=None, chapter=None, verse=None):
        match format: 
            case 'txt':
                self.print_wall_of_text(verse_records, verse_numbers)

            case 'md':
                self.print_markdown_excerpt(verse_records, book, chapter, verse)

    def print_book(self, book, format, verse_numbers):
        cursor = self.get_bible_cursor()
        book = self.get_book_from_abbreviation(book)
        params = {'book': book}
    
        cursor.execute("""
        SELECT verse, text FROM verses
        JOIN books ON verses.book_id = books.id
        WHERE books.name = :book
        """, params)

        verse_records = cursor.fetchall()
        
        self.print_passage_by_format(
            format,
            verse_records,
            verse_numbers=verse_numbers,
            book=book
        )

    # TODO: Validate chapter?
    def print_chapter(self, book, chapter, format, verse_numbers):
        cursor = self.get_bible_cursor()
        book = self.get_book_from_abbreviation(book)
        params = {'book': book, 'chapter': chapter}
        
        cursor.execute("""
        SELECT verse, text FROM verses
        JOIN books ON verses.book_id = books.id
        WHERE books.name = :book
        AND chapter = :chapter
        """, params)

        verse_records = cursor.fetchall()
        
        if len(verse_records) == 0:
            msg = f"Invalid chapter: {book} {chapter}."
            print(msg)
        
        else:
            self.print_passage_by_format(
                format,
                verse_records,
                verse_numbers=verse_numbers,
                book=book,
                chapter=chapter
            )

    # TODO: Validate chapter?
    def print_verse(self, book, chapter, verse, format, verse_numbers):
        cursor = self.get_bible_cursor()
        book = self.get_book_from_abbreviation(book)
        params = {
            'book': book,
            'chapter': chapter,
            'verse': verse
        }
        
        cursor.execute("""
        SELECT verse, text FROM verses
        JOIN books ON verses.book_id = books.id
        WHERE books.name = :book
        AND chapter = :chapter
        AND verse = :verse
        """, params)

        verse_records = cursor.fetchall()
        
        if len(verse_records) == 0:
            msg = f"Invalid verse: {book} {chapter}:{verse}."
            print(msg)
        
        else:
            self.print_passage_by_format(
                format,
                verse_records,
                verse_numbers=verse_numbers,
                book=book,
                chapter=chapter,
                verse=verse
            )

    # TODO: Validate chapter?
    def print_verses(self, book, chapter, verse, format, verse_numbers):
        """
        Print a range of verses, eg. 5-7. 
        """
        cursor = self.get_bible_cursor()
        book = self.get_book_from_abbreviation(book)
        verse_start, verse_end = parse_verses_str(verse)
        
        params = {
            'book': book,
            'chapter': chapter,
            'verse_start': verse_start,
            'verse_end': verse_end,
        }
        
        cursor.execute("""
        SELECT verse, text FROM verses
        JOIN books ON verses.book_id = books.id
        WHERE books.name = :book
        AND chapter = :chapter
        AND verse BETWEEN :verse_start AND :verse_end
        """, params)

        verse_records = cursor.fetchall()
        
        if len(verse_records) == 0:
            msg = (
                f"Invalid verses: {book} "
                f"{chapter}:{verse_start}-{verse_end}."
            )
            print(msg)
        
        else:
            self.print_passage_by_format(
                format,
                verse_records,
                verse_numbers=verse_numbers,
                book=book,
                chapter=chapter,
                verse=verse
            )
    
    # TODO: Output txt, markdown table, csv format
    def search_bible(self, phrase):
        cursor = self.get_bible_cursor()
        
        cursor.execute("""
        SELECT books.name AS book, chapter, verse, text FROM verses
        JOIN books ON verses.book_id = books.id
        WHERE verses.text LIKE ?;
        """, (f"%{phrase}%",))
        
        verse_records = cursor.fetchall()
        
        print(f"{len(verse_records)} occurrences of '{phrase}' in the {self.translation} Bible:\n___\n")
        
        for verse in verse_records:
            # TODO: Use/rename create_link_label? (omit translation)
            print(f"{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n")
    
    def search_testament(self, phrase, testament):
        cursor = self.get_bible_cursor()

        new_testament = [
            'Matthew',
            'Mark',
            'Luke',
            'John',
            'Acts',
            'Romans',
            'I Corinthians',
            'II Corinthians',
            'Galatians',
            'Ephesians',
            'Philippians',
            'Colossians',
            'I Thessalonians',
            'II Thessalonians',
            'I Timothy',
            'II Timothy',
            'Titus',
            'Philemon',
            'Hebrews',
            'James',
            'I Peter',
            'II Peter',
            'I John',
            'II John',
            'III John',
            'Jude',
            'Revelation of John'
        ]

        nt_sql_list = list_to_sql(new_testament)

        if testament == 'nt':

            sql = f"""
            SELECT books.name AS book, chapter, verse, text FROM verses
            JOIN books ON verses.book_id = books.id
            WHERE verses.text LIKE ?
            AND books.name IN {nt_sql_list};
            """
        
        elif testament == 'ot': 

            sql = f"""
            SELECT books.name AS book, chapter, verse, text FROM verses
            JOIN books ON verses.book_id = books.id
            WHERE verses.text LIKE ?
            AND books.name NOT IN {nt_sql_list};
            """
        
        else:
            sys.exit(f"Invalid {testament=}.")
        
        # Bind phrase since it's user input
        cursor.execute(sql, (f"%{phrase}%",))
        verse_records = cursor.fetchall()

        testament = 'New Testament' if testament == 'nt' else 'Old Testament'
        print(f"{len(verse_records)} occurrences of '{phrase}' in the {testament} ({self.translation}):\n___\n")
        
        for verse in verse_records:
            # TODO: Use/rename create_link_label? (omit translation)
            print(f"{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n")
    
    def search_book(self, phrase, book):
        cursor = self.get_bible_cursor()

        book = self.get_book_from_abbreviation(book)
        params = {
            'phrase': f"%{phrase}%",
            'book': book,
            }
        
        cursor.execute("""
        SELECT books.name AS book, chapter, verse, text FROM verses
        JOIN books ON verses.book_id = books.id
        WHERE verses.text LIKE :phrase
        AND book = :book;
        """, params)
        
        verse_records = cursor.fetchall()
        
        print(f"{len(verse_records)} occurrences of '{phrase}' in {book} ({self.translation}):\n___\n")
        
        for verse in verse_records:
            # TODO: Use/rename create_link_label? (omit translation)
            print(f"{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n")

    # TODO: Validate chapter?
    def search_chapter(self, phrase, book, chapter):
        cursor = self.get_bible_cursor()

        book = self.get_book_from_abbreviation(book)
        params = {
            'phrase': f"%{phrase}%",
            'book': book,
            'chapter': chapter,
            }
        
        cursor.execute("""
        SELECT books.name AS book, chapter, verse, text FROM verses
        JOIN books ON verses.book_id = books.id
        WHERE verses.text LIKE :phrase
        AND book = :book
        AND chapter = :chapter;
        """, params)
        
        verse_records = cursor.fetchall()
        
        print(f"{len(verse_records)} occurrences of '{phrase}' in {book} {chapter} ({self.translation}):\n___\n")
        
        for verse in verse_records:
            # TODO: Use/rename create_link_label? (omit translation)
            print(f"{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n")
